#include "GPUFontLayer.h"
#include "Atlas.h"
#include "AtlasLayerCombo.h"
#include "Utils.h"
#include "font.hpp"
#include "fwd.hpp"
#include "ofGraphics.h"
#include "ofUtils.h"
#include "ofxProfiler.h"

namespace ofxVariableLab {

void GPUFontLayer::setup(const LayerSettings & settings){
    this->settings = settings;
    if(id == ""){
        id = "layer-" + ofToString(Layer::n_layers);
        n_layers++;
    }
    innerNode.setParent(outerNode);
}

void GPUFontLayer::update(){
    OFX_PROFILER_FUNCTION();
    // TODO: figure out why this makes it slow
    //if(propsBuffer.size() > 0){
    //propsBuffer.push_front(lastProps);
    //propsBuffer.pop_back();
    //const Props & props = lastProps;
    //for(int i = 0; i < props.text.size(); i++){
    //const Props & vProps = getProps(i * props.letterDelay);
    //ofxGPUFont::GlyphAppearance & glyphAppearance = variationTextAppearance[i];
    //glyphAppearance.fontSize_px = vProps.fontSize_px;
    //variationTextAppearance[i].color = vProps.color;
    //}
    //}
}

void GPUFontLayer::draw(glm::vec3 position){
}
void GPUFontLayer::drawCharacter(const char character,
                                 glm::vec3 position,
                                 glm::vec4 color,
                                 float rotation,
                                 float scale,
                                 ofxVariableLab::FontVariation fontVariation){
}
const LayerType & GPUFontLayer::getType() const {
    return type;
}
void GPUFontLayer::setVFlip(const VFlipState vFlip){
    this->vFlip = vFlip;
}

bool GPUFontLayer::isDirtyDirty() const {
    return isDirty;
}

void GPUFontLayer::setDirtyDirty(bool dirtyDirty){
    isDirty = dirtyDirty;
}

bool GPUFontLayer::wantsNewMom(){
    return notHappyWithMom;
}
void GPUFontLayer::isWithNewMom(){
    notHappyWithMom = false;
}
const ComboIdentifier & GPUFontLayer::getMomsComboIdentifier() const {
    return momsComboIdentifier;
}
void GPUFontLayer::setMomsComboIdentifier(const ComboIdentifier & identifier){
    momsComboIdentifier = identifier;
}

void GPUFontLayer::setProps(const Props & props){
    OFX_PROFILER_FUNCTION();
    bool propsBufferEmpty = propsBuffer.size() == 0;
    bool setAppearanceAlready = false;
    if(propsBufferEmpty || lastProps.text != props.text
       || lastProps.fontVariations != props.fontVariations){
        setAppearanceAlready = true;
        OFX_PROFILER_SCOPE("variations");
        setDirtyDirty(true);
        variationText.clear();
        variationText.resize(props.text.size());
        variationTextAppearance.clear();
        variationTextAppearance.resize(props.text.size());
        int i = 0;
        for(const char c : props.text){
            OFX_PROFILER_SCOPE("char");
            const Props & vProps = propsBufferEmpty ? props : getProps(i * props.letterDelay);
            ofxGPUFont::GlyphIdentity glyphIdentity;
            ofxGPUFont::GlyphAppearance glyphAppearance;
            glyphIdentity.charcode = ofxGPUFont::decodeCharcode(&c);
            glyphAppearance.charcode = glyphIdentity.charcode;
            glyphAppearance.color = vProps.color;
            glyphAppearance.fontSize_px = vProps.fontSize_px;
            for(const auto & fv : vProps.fontVariations){
                glyphIdentity.coords.push_back(
                    std::move(ofxGPUFont::float2f16dot16(round(fv.value * 0.1) * 10))     // convert to FT
                    // NOTE: we are also rounding here, since in practice a variation
                    // of 100.0 will not look much different than 100.124
                    );
            }
            variationText[i] = std::move(glyphIdentity);
            variationTextAppearance[i] = std::move(glyphAppearance);
            i++;
        }
    }
    if(!setAppearanceAlready
       && (lastProps.fontSize_px != props.fontSize_px
           || lastProps.color != props.color)){
        for(int i = 0; i < props.text.size(); i++){
            const Props & vProps = propsBufferEmpty ? props : getProps(i * props.letterDelay);
            ofxGPUFont::GlyphAppearance & glyphAppearance = variationTextAppearance[i];
            glyphAppearance.fontSize_px = vProps.fontSize_px;
            variationTextAppearance[i].color = vProps.color;
        }
    }
    if(lastProps.fontPath != props.fontPath){
        notHappyWithMom = true;
    }

    int maxPropsBufferSize = max(0, int(props.letterDelay * settings.letterDelayRatio * props.text.size()));
    if(maxPropsBufferSize > 0){
        propsBuffer.push_front(props);
    }
    while(propsBuffer.size() > maxPropsBufferSize){
        propsBuffer.pop_back();
    }
    lastProps = props;
}
const Layer::Props & GPUFontLayer::getProps(float delay_seconds) const {
    if(delay_seconds != 0 && propsBuffer.size() > 0){
        int index = max(0, min(int(propsBuffer.size() - 1), int(delay_seconds * settings.letterDelayRatio)));
        return propsBuffer[index]; // gets newest
    }else{
        return lastProps;
    }
}
ofNode & GPUFontLayer::getOuterNode(){
    return outerNode;
}
ofNode & GPUFontLayer::getInnerNode(){
    return innerNode;
}
void GPUFontLayer::clearPropsBuffer(){
    propsBuffer.clear();
}
void GPUFontLayer::setId(const LayerID & id){
    this->id = id;
}
const LayerID & GPUFontLayer::getId(){
    return id;
}
const vector <ofxGPUFont::GlyphIdentity> & GPUFontLayer::getVariationText(){
    return variationText;
}
const vector <ofxGPUFont::GlyphAppearance> & GPUFontLayer::getVariationTextAppearance(){
    return variationTextAppearance;
}
}
